package net.christopherschultz.util.zipcode;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Serializable;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.MissingResourceException;

/**
 * <p>A registry of US postal codes.</p>
 *
 * <p>
 *     The methods in this class are not thread safe, so intermixing
 *     calls to {@link #add} with the various forms of <code>get</code>
 *     should be synchronized. Usually, the registry is initialized once
 *     and then all access is read-only; hence, this class provides no
 *     synchronization of its own.
 * </p>
 *
 * @author Chris Schultz
 * @version $Revision$ $Date$
 */
public class ZipCodeRegistry
    implements Serializable
{
    /**
     * <p>Map of zip codes to ZipCode objects.</p>
     *
     * <p>The key is an Integer (i.e. 20005) and the values are ZipCode
     * objects.</p>
     */
    private HashMap _zipCodes;

    /**
     * <p>Creates a ZipCodeRegistry object with no ZipCodes.</p>
     */
    public ZipCodeRegistry()
    {
	_zipCodes = new HashMap();
    }

    /**
     * <p>Adds a ZipCode to the registry.</p>
     *
     * @param zipCode The ZipCode to add to the registry.
     */
    public void add(ZipCode zipCode)
    {
	_zipCodes.put(new Integer(zipCode.getZipCode()), zipCode);
    }

    /**
     * <p>Gets the ZipCode which corresponds to the specified zip code.</p>
     *
     * @param zipCode The 5-digit zip code to lookup.
     *
     * @return The ZipCode object that corresponds to the specified zip code,
     *         or <code>null</code> if there is no corresponding ZipCode
     *         object for the specified zip code.
     *
     * @see #getNearestMatch
     */
    public ZipCode getZipCode(int zipCode)
    {
	return (ZipCode)_zipCodes.get(new Integer(zipCode));
    }

    /**
     * <p>Gets the ZipCode which is the closest numerically to the specified
     * zip code.</p>
     *
     * <p>Numerically looking up zip codes is rather inaccurate, but if you
     * have a zip code which is not in the registry, this may be the only
     * way to find one.</p>
     *
     * @param zipCode The 5-digit zip code to lookup.
     *
     * @return The ZipCode object that corresponds to the zip code closest
     *         numerically to the specified zip code.
     */
    public ZipCode getNearestMatch(int zipCode)
    {
	ZipCode zc = null;
	int change = 1;

	while(null == zc)
	{
	    Integer code = new Integer(zipCode);
	    zc = (ZipCode)_zipCodes.get(code);

	    if(null != zc)
		return zc;

	    zipCode += change;

	    if(change < 0)
		change = -(change - 1);
	    else
		change = -(change + 1);
	}

	return zc;
    }

    /**
     * <p>Gets the ZipCode that is closest geographically from the specified
     * ZipCode.</p>
     *
     * @param target The ZipCode to come closest to.
     *
     * @return The ZipCode which is closest geographically to the specified
     *         ZipCode.
     */
    public ZipCode getClosestZipCode(ZipCode target)
    {
	double arc = Math.PI; // You'd better be closer than that!
	ZipCode closest = null;

	for(Iterator i=_zipCodes.values().iterator(); i.hasNext(); )
	{
	    ZipCode zipCode = (ZipCode)i.next();

	    if(zipCode.equals(target))
		continue;

	    double theta = zipCode.getArcLength(target);

	    if(theta < arc)
	    {
		closest = zipCode;

		arc = theta;
	    }
	}

	return closest;
    }

    /**
     * <p>Gets a list of ZipCodes that are within the specified number of
     * miles.</p>
     *
     * @param target The ZipCode which should be the center of the search.
     * @param miles The maximum distance (in miles) a ZipCode may be from the
     *        target.
     *
     * @return A list of ZipCode objects each of which has a distance
     *         less than or equal to the specified mileage.
     */
    public List getZipCodesWithinMiles(ZipCode target, double miles)
    {
	ArrayList zipCodes = new ArrayList();

	for(Iterator i=_zipCodes.values().iterator(); i.hasNext(); )
	{
	    ZipCode zipCode = (ZipCode)i.next();

	    if(miles >= zipCode.getDistanceInMiles(target))
		zipCodes.add(zipCode);
	}

	return zipCodes;
    }

    /**
     * <p>Gets a list of ZipCodes that are within the specified number of
     * kilometers.</p>
     *
     * @param target The ZipCode which should be the center of the search.
     * @param km The maximum distance (in kilometers) a ZipCode may be from the
     *        target.
     *
     * @return A list of ZipCode objects each of which has a distance
     *         less than or equal to the specified distance in kilometers.
     */
    public List getZipCodesWithinKilometers(ZipCode target, double km)
    {
	ArrayList zipCodes = new ArrayList();

	for(Iterator i=_zipCodes.values().iterator(); i.hasNext(); )
	{
	    ZipCode zipCode = (ZipCode)i.next();

	    if(km >= zipCode.getDistanceInKilometers(target))
		zipCodes.add(zipCode);
	}

	return zipCodes;
    }

    /**
     * <p>Gets the default instance of a ZipCodeRegistry.</p>
     *
     * <p>This ZipCodeRegistry comes with a zip code location file that it will
     * load for itself.</p>
     *
     * @return A ZipCodeRegistry containing ZipCode objects for all zip codes.
     *
     * @throws IOException If there is a problem loading the zip code registry.
     */
    public static ZipCodeRegistry getInstance()
	throws IOException
    {
	return load("zipcodes.txt");
    }

    /**
     * <p>Loads a ZipCodeRegistry from the specified file.</p>
     *
     * <p>The file should be text with each line like this:</p>
     *
     * <pre>
     *         byte
     *         0123456789 123456789 12345
     * content [ZIP][LATITUDE][LONGITUDE]
     * </pre>
     *
     * <p>...where [ZIP] is the 5-digit zip-code and [LATITUDE] and [LONGITUDE]
     * are the latitude and longitude, respectively, of the post office
     * serving that zip code.</p>
     *
     * <p>The latitude and longitude may be of arbitrary precision, and may
     * begin with a - or + sign. Note that there is no whitespace of any kind
     * necessary. Whitespace must be used to pad any missing digits since
     * the byte count is significant.</p>
     *
     * <p>Any lines of text starting with a hash mark (#) will be ignored.</p>
     *
     * @param resourceName The name of the resource containing the
     *        properly serialized form of the ZipCodeRegistry (see above).
     *
     * @return A ZipCodeRegistry loaded from the specified file.
     *
     * @throws IOException If there is a problem reading the resource.
     */
    public static ZipCodeRegistry load(String resourceName)
	throws IOException
    {
	BufferedReader in = null;

	try
	{
	    java.io.InputStream stream = ZipCodeRegistry.class.getClassLoader().getResourceAsStream(resourceName);

	    if(null == stream)
		throw new java.io.FileNotFoundException(resourceName);

	    in = new BufferedReader(new InputStreamReader(stream));

	    ZipCodeRegistry registry = new ZipCodeRegistry();
	    String line;
	    registry = new ZipCodeRegistry();
	    while(null != (line = in.readLine()))
	    {
		if(line.startsWith("#"))
		    continue;

		String zip_code   = line.substring(0, 5);
		String latitude   = line.substring(5, 15);
		String longitude  = line.substring(15);

		Integer zipcode = Integer.valueOf(zip_code);
		ZipCode zip = new ZipCode(zipcode.intValue(),
					  Double.parseDouble(latitude),
					  Double.parseDouble(longitude));

		registry.add(zip);
	    }

	    return(registry);
	}
	finally
	{
	    if(null != in)
	    { try { in.close(); } catch (IOException ioe)
	    { ioe.printStackTrace(); }
	    }
	}
    }
}
